# music_sync - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""sync_music - Sync music library to external device."""

import os
import codecs
import logging
import argparse
import configparser
import sys

from multiprocessing import Pool

import pbr.version

from . import util
from .hashdb import HashDb
from .actions import Copy
from .actions import Skip
from .transcode import Transcode

__version__ = pbr.version.VersionInfo('kasserver').release_string()

logger = util.LogStyleAdapter(  # pylint: disable=invalid-name
    logging.getLogger(__name__))


class SyncMusic():
    """sync_music - Sync music library to external device."""

    def __init__(self, args):
        """Initialize SyncMusic."""
        logger.info(__doc__)
        logger.info("")
        self._args = args
        self._hashdb = HashDb(os.path.join(args.audio_dest, 'sync_music.db'))
        logger.info("Settings:")
        logger.info(" - audio-src:  {}".format(args.audio_src))
        logger.info(" - audio-dest: {}".format(args.audio_dest))
        if args.playlist_src:
            logger.info(" - playlist-src: {}".format(args.playlist_src))
        logger.info(" - mode: {}".format(args.mode))
        logger.info("")
        self._action_copy = Copy()
        self._action_skip = Skip()
        self._action_transcode = Transcode(
            mode=self._args.mode,
            replaygain_preamp_gain=self._args.replaygain_preamp_gain,
            transcode=not self._args.disable_file_processing,
            copy_tags=not self._args.disable_tag_processing,
            albumartist_artist_hack=self._args.albumartist_artist_hack,
            albumartist_composer_hack=self._args.albumartist_composer_hack,
            discnumber_hack=self._args.discnumber_hack,
            tracknumber_hack=self._args.tracknumber_hack)

    def _process_file(self, current_file):
        """Process single file.

        :param current_file: tuple:
            (file_index, total_files, in_filename, action)
        """
        file_index, total_files, in_filename, action = current_file
        out_filename = action.get_out_filename(in_filename)
        if out_filename is not None:
            out_filename = util.correct_path_fat32(out_filename)
            logger.info("{:04}/{:04}: {} {} to {}",
                        file_index, total_files, action.name,
                        in_filename, out_filename)
        else:
            logger.info("{:04}/{:04}: {} {}",
                        file_index, total_files, action.name, in_filename)
            return None

        in_filepath = os.path.join(self._args.audio_src, in_filename)
        out_filepath = os.path.join(self._args.audio_dest, out_filename)

        # Calculate hash to see if the input file has changed
        hash_current = self._hashdb.get_hash(in_filepath)
        hash_database = None
        if in_filename in self._hashdb.database:
            hash_database = self._hashdb.database[in_filename][1]

        if (self._args.force or hash_database is None
                or hash_database != hash_current
                or not os.path.exists(out_filepath)):
            util.ensure_directory_exists(os.path.dirname(out_filepath))
            try:
                action.execute(in_filepath, out_filepath)
            except IOError as err:
                logger.error("Error: {}", err)
                return None
            return (in_filename, out_filename, hash_current)
        logger.info("Skipping up to date file")
        return None

    def _get_file_action(self, in_filename):
        """Determine the action for the given file."""
        extension = os.path.splitext(in_filename)[1]
        if extension in ['.flac', '.ogg', '.mp3']:
            if self._args.mode == 'copy':
                return self._action_copy
            return self._action_transcode
        if in_filename.endswith('folder.jpg'):
            return self._action_copy
        return self._action_skip

    def _clean_up_missing_files(self):
        """Remove files in the destination, where the source file doesn't
           exist anymore.
        """
        logger.info("Cleaning up missing files")
        files = [(k, v[0]) for k, v in self._hashdb.database.items()]
        for in_filename, out_filename in files:
            in_filepath = os.path.join(self._args.audio_src, in_filename)
            out_filepath = os.path.join(self._args.audio_dest, out_filename)

            if not os.path.exists(in_filepath):
                if os.path.exists(out_filepath):
                    if (self._args.batch or util.query_yes_no(
                            "File {} does not exist, do you want to remove {}"
                            .format(in_filename, out_filename))):
                        try:
                            os.remove(out_filepath)
                        except OSError as err:
                            logger.error("Error: Failed to remove file {}",
                                         err)
                if not os.path.exists(out_filepath):
                    del self._hashdb.database[in_filename]

    def _clean_up_empty_directories(self):
        """Remove empty directories in the destination."""
        logger.info("Cleaning up empty directories")
        util.delete_empty_directories(self._args.audio_dest)

    def sync_audio(self):
        """Sync audio."""
        self._hashdb.load()

        # Create a list of all tracks ordered by their last modified time stamp
        files = [(f, self._get_file_action(f),
                  os.path.getmtime(os.path.join(self._args.audio_src, f)))
                 for f in util.list_all_files(self._args.audio_src)]
        files = [(index, len(files), f[0], f[1])
                 for index, f in enumerate(files, 1)]
        if not files:
            raise FileNotFoundError("No input files")

        # Cleanup files that does not exist any more
        self._clean_up_missing_files()
        self._clean_up_empty_directories()

        # Do the work
        logger.info("Starting actions")
        try:
            if self._args.jobs == 1:
                # pool.map doesn't might not show all exceptions
                file_hashes = []
                for current_file in files:
                    file_hashes.append(self._process_file(current_file))
            else:
                pool = Pool(processes=self._args.jobs)
                file_hashes = pool.map(self._process_file, files)
        except:  # noqa, pylint: disable=bare-except
            logger.error(">>> traceback <<<")
            logger.exception("Exception")
            logger.error(">>> end of traceback <<<")

        # Store new hashes in the database
        for file_hash in file_hashes:
            if file_hash is not None:
                self._hashdb.database[file_hash[0]] = \
                    (file_hash[1], file_hash[2])
        self._hashdb.store()

    def sync_playlists(self):
        """Sync m3u playlists."""
        for dirpath, _, filenames in os.walk(self._args.playlist_src):
            relpath = os.path.relpath(dirpath, self._args.playlist_src)
            for filename in filenames:
                if os.path.splitext(filename)[1] == '.m3u':
                    try:
                        self._sync_playlist(
                            os.path.normpath(
                                os.path.join(relpath, filename)))
                    except IOError as err:
                        logger.error("Error: {}", err)

    def _sync_playlist(self, filename):
        """Sync playlist."""
        logger.info("Syncing playlist {}", filename)
        srcpath = os.path.join(self._args.playlist_src, filename)
        destpath = os.path.join(self._args.audio_dest, filename)

        if os.path.exists(destpath):
            os.remove(destpath)

        # Copy file
        in_file = codecs.open(srcpath, 'r', encoding='windows-1252')
        out_file = codecs.open(destpath, 'w', encoding='windows-1252')
        for line in in_file.read().splitlines():
            if not line.startswith('#EXT'):
                in_filename = line
                try:
                    while True:
                        if in_filename in self._hashdb.database:
                            line = self._hashdb.database[in_filename][0]
                            line = line.replace('/', '\\')
                            break
                        in_filename = in_filename.split('/', 1)[1]
                except IndexError:
                    logger.warning("File does not exist: {}", line)
                    continue
            line = line + '\r\n'
            out_file.write(line)
        in_file.close()
        out_file.close()


def load_settings(arguments=None):  # pylint: disable=too-many-locals
    """Load settings."""
    # ArgumentParser 1: Get config file (disable help)
    config_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)
    config_parser.add_argument('-c', '--config-file',
                               help='specify config file', metavar='FILE')
    args, remaining_argv = config_parser.parse_known_args(arguments)

    # Read default settings from config file
    if args.config_file is None:
        args.config_file = util.makepath('~/.sync_music')
    config = configparser.ConfigParser()
    config.read([args.config_file])
    try:
        defaults = dict(config.items("Defaults"))
    except configparser.NoSectionError:
        defaults = {}

    # ArgumentParser 2: Get rest of the arguments
    parser = argparse.ArgumentParser(parents=[config_parser])
    parser.set_defaults(**defaults)
    parser.add_argument(
        '-v', '--version', action='version',
        version='%(prog)s {}'.format(__version__))
    parser.add_argument(
        '-b', '--batch', action='store_true', help="batch mode, no user input")
    parser.add_argument(
        '-o', '--logfile', type=str, default='./sync_music.log',
        help="write log output to file")

    parser_paths = parser.add_argument_group("Paths")
    parser_paths.add_argument(
        '--audio-src', type=str, required='audio_src' not in defaults,
        help="folder containing the audio sources")
    parser_paths.add_argument(
        '--audio-dest', type=str, required='audio_dest' not in defaults,
        help="target directory for converted files")
    parser_paths.add_argument(
        '--playlist-src', type=str,
        help='folder containing the source playlists')

    # Audio sync options
    parser_audio = parser.add_argument_group("Transcoding options")
    parser_audio.add_argument(
        '--mode',
        choices=['auto', 'transcode', 'replaygain', 'replaygain-album',
                 'copy'],
        default='auto',
        help="auto: copy MP3s, transcode others and adapt tags (default); "
             "transcode: transcode all files and adapt tags (slow); "
             "replaygain: transcode all files, apply ReplayGain track based "
             "normalization and adapt tags (slow), "
             "replaygain-album: transcode all files, apply ReplayGain album "
             "based normalization and adapt tags (slow), "
             "copy: copy all files, leave tags untouched (implies "
             "--disable-tag-processing)")
    parser_audio.add_argument(
        '--replaygain-preamp-gain', type=float,
        default=4.0,
        help="modify ReplayGain pre-amp gain if transcoded files are "
             "too quiet or too loud (default +4.0 as many players are "
             "calibrated for higher volume)")
    parser_audio.add_argument(
        '--disable-file-processing', action='store_true',
        help="disable processing files, update tags "
             "(if not explicitly disabled)")
    parser_audio.add_argument(
        '--disable-tag-processing', action='store_true',
        help="disable processing tags, update files "
             "(if not explicitly disabled)")
    parser_audio.add_argument(
        '-f', '--force', action='store_true',
        help="rerun action even if the source file has not changed")
    parser_audio.add_argument(
        '-j', '--jobs', type=int, default=4, help="number of parallel jobs")

    # Optons for action transcode
    parser_hacks = parser.add_argument_group(
        "Hacks", "Modify target files to work around player shortcomings")
    parser_hacks.add_argument(
        '--albumartist-artist-hack', action='store_true',
        help="write album artist into artist field")
    parser_hacks.add_argument(
        '--albumartist-composer-hack', action='store_true',
        help="write album artist into composer field")
    parser_hacks.add_argument(
        '--discnumber-hack', action='store_true',
        help="extend album field by disc number")
    parser_hacks.add_argument(
        '--tracknumber-hack', action='store_true',
        help="remove track total from track number")

    # Parse
    settings = parser.parse_args(remaining_argv)

    # Check required arguments and make absolute paths
    try:
        if settings.mode == 'copy' and (settings.albumartist_artist_hack or
                                        settings.albumartist_composer_hack or
                                        settings.discnumber_hack or
                                        settings.tracknumber_hack):
            parser.error("hacks cannot be used in copy mode")
        paths = ['audio_src', 'audio_dest']
        if settings.playlist_src is not None:
            paths.append('playlist_src')
        settings_dict = vars(settings)
        util.ensure_directory_exists(
            util.makepath(settings_dict['audio_dest']))
        for path in paths:
            settings_dict[path] = util.makepath(settings_dict[path])
            if not os.path.isdir(settings_dict[path]):
                raise IOError("{} is not a directory".format(
                    settings_dict[path]))
    except IOError as err:
        parser.error(err)

    return settings


def main():  # pragma: no cover
    """sync_music - Sync music library to external device."""
    args = load_settings()

    rootlogger = logging.getLogger()
    rootlogger.setLevel(logging.DEBUG)

    consolelogger = logging.StreamHandler(sys.stdout)
    consolelogger.setLevel(logging.INFO)
    consolelogger.setFormatter(logging.Formatter("{message}", style='{'))
    rootlogger.addHandler(consolelogger)

    if args.logfile:
        filelogger = logging.FileHandler(args.logfile, 'w')
        filelogger.setLevel(logging.INFO)
        filelogger.setFormatter(logging.Formatter(
            "{asctime} {levelname:8s} {message}", style='{'))
        rootlogger.addHandler(filelogger)

    sync_music = SyncMusic(args)

    if not args.batch and not util.query_yes_no("Do you want to continue?"):
        sys.exit(1)

    try:
        sync_music.sync_audio()
    except FileNotFoundError as err:
        logger.critical("Failed to sync music {}", err)
        sys.exit(1)

    if args.playlist_src:
        sync_music.sync_playlists()



# sync_music - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Basic actions."""

import shutil


class Copy:
    """Copy action simply copies file."""

    def __init__(self):
        self.name = "Copying"

    @classmethod
    def get_out_filename(cls, path):
        """Determine output file path."""
        return path

    @classmethod
    def execute(cls, in_filepath, out_filepath):
        """Executes action."""
        shutil.copy(in_filepath, out_filepath)


class Skip:
    """Skip action does nothing."""

    def __init__(self):
        self.name = "Skipping"

    @classmethod
    def get_out_filename(cls, _):
        """Determine output file path."""

    @classmethod
    def execute(cls, in_filepath, out_filepath):  # pragma: no cover
        """Executes action."""



# sync_music - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""sync_music - Sync music library to external device."""

import sys

from platform import python_version
from pkg_resources import parse_version


def verify_interpreter_version():
    """Verify the Pyhton interpreter version."""
    minversion = '3.5'
    version = python_version()
    if parse_version(version) < parse_version(minversion):  # pragma: no cover
        sys.stdout.write("Incompatible Python version, minimum supported "
                         "version {}, found version {}\n".format(
                             minversion, version))
        sys.exit(1)


verify_interpreter_version()



# music_sync - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Tests the transcoding implementation."""

import os
import shutil

import pytest

from sync_music.sync_music import Transcode


class TestTranscode():
    """Tests the transcoding implementation."""

    in_filename_flac = 'withtags.flac'
    in_filename_flacall = 'withalltags.flac'
    in_filename_flacempty = 'stripped.flac'
    in_filename_ogg = 'withtags.ogg'
    in_filename_oggall = 'withalltags.ogg'
    in_filename_oggempty = 'stripped.ogg'
    in_filename_mp3 = 'withtags.mp3'
    in_filename_mp3all = 'withalltags.mp3'
    in_filename_mp3empty = 'stripped.mp3'
    in_filename_aiff = 'withtags.aiff'
    out_filename = 'withtags.mp3'
    img_filename = 'folder.jpg'
    input_path = 'tests/reference_data/audiofiles'
    output_path = None

    @pytest.fixture(autouse=True)
    def init_output_path(self, tmpdir):
        """Setup temporary output directory."""
        self.output_path = str(tmpdir)

    def test_filename(self):
        """Tests retrieving the output file name."""
        transcode = Transcode()
        out_filename = transcode.get_out_filename(self.in_filename_flac)
        assert self.out_filename == out_filename

    def execute_transcode(self, transcode, in_filename=in_filename_flac,
                          out_filename=out_filename):
        """Helper method to run transcoding tests."""
        transcode.execute(
            os.path.join(self.input_path, in_filename),
            os.path.join(self.output_path, out_filename))

    def test_transcode_default(self):
        """Test transcoding with default options."""
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_flac)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_flacall)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_flacempty)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_ogg)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_oggall)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_oggempty)

    def test_transcode_copy(self):
        """Tests transcoding with copying instead of transcoding."""
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_mp3)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_mp3all)
        self.execute_transcode(Transcode(),
                               in_filename=self.in_filename_mp3empty)

    def test_transcode_transcode(self):
        """Tests transcoding with forced transcode."""
        self.execute_transcode(Transcode(mode='transcode'),
                               in_filename=self.in_filename_mp3)

    def test_transcode_replaygain(self):
        """Tests transcoding with ReplayGain (track based)."""
        self.execute_transcode(Transcode(mode='replaygain'),
                               in_filename=self.in_filename_mp3all)
        self.execute_transcode(Transcode(mode='replaygain'),
                               in_filename=self.in_filename_mp3)

    def test_transcode_replaygainalbum(self):
        """Tests transcoding with ReplayGain (album based)."""
        self.execute_transcode(Transcode(mode='replaygain-album',
                                         replaygain_preamp_gain=10.0),
                               in_filename=self.in_filename_mp3)

    def test_transcode_folderimage(self):
        """Tests transcoding without folder image."""
        # Copy input file to a folder without folder.jpg (output folder)
        in_filename = os.path.join(self.output_path, self.in_filename_flac)
        shutil.copy(
            os.path.join(self.input_path, self.in_filename_flac),
            in_filename)
        self.execute_transcode(Transcode(), in_filename=in_filename)

    def test_transcode_artist_hack(self):
        """Tests transcoding with albumartist to artist hack enabled."""
        self.execute_transcode(Transcode(albumartist_artist_hack=True))
        self.execute_transcode(Transcode(albumartist_artist_hack=True),
                               in_filename=self.in_filename_mp3empty)

    def test_transcode_composer_hack(self):
        """Tests transcoding with albumartist to composer hack enabled."""
        self.execute_transcode(Transcode(albumartist_composer_hack=True))
        self.execute_transcode(Transcode(albumartist_composer_hack=True),
                               in_filename=self.in_filename_mp3empty)

    def test_transcode_discnumber_hack(self):
        """Tests transcoding with disc number hack enabled."""
        self.execute_transcode(Transcode(discnumber_hack=True))
        self.execute_transcode(Transcode(discnumber_hack=True),
                               in_filename=self.in_filename_mp3all)

    def test_transcode_tracknumber_hack(self):
        """Tests transcoding with track number hack enabled."""
        self.execute_transcode(Transcode(tracknumber_hack=True))
        self.execute_transcode(Transcode(tracknumber_hack=True),
                               in_filename=self.in_filename_mp3empty)
        self.execute_transcode(Transcode(tracknumber_hack=True),
                               in_filename='brokentag_tracknumber.mp3')

    def test_transcodeerror_transcode(self):
        """Tests transcoding failure."""
        # IOError is raised on audiotools.EncodingError. The easiest way that
        # leads into this exception is writing to a non writable path.
        with pytest.raises(IOError):
            self.execute_transcode(Transcode(), out_filename='/')

    def test_transcodeerror_copytags(self):
        """Tests copying tag failure."""
        # Copy tags expects the output file to be an MP3 file.
        # Don't transcode as this would rewrite the output file.
        shutil.copy(
            os.path.join(self.input_path, self.img_filename),
            os.path.join(self.output_path, self.out_filename))
        with pytest.raises(IOError):
            self.execute_transcode(Transcode(mode='copy'))

    def test_transcodingerror_format(self):
        """Tests transcoding a non supported format."""
        with pytest.raises(IOError):
            self.execute_transcode(Transcode(),
                                   in_filename=self.in_filename_aiff)



# sync_music - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Utilities."""

import logging
import os
import sys
import re


# Utility classes that allow using the built-in logging facilities with
# the newer string.format style instead of the '%' style.
# From: https://docs.python.org/3/howto/logging-cookbook.html

class LogBraceString:  # pylint: disable=too-few-public-methods
    """Log message that supports string.format()."""
    def __init__(self, fmt, args):
        self.fmt = fmt
        self.args = args

    def __str__(self):
        return self.fmt.format(*self.args)


class LogStyleAdapter(logging.LoggerAdapter):
    """Logging StyleAdapter that supports string.format()."""
    def __init__(self, logger_instance, extra=None):
        super(LogStyleAdapter, self).__init__(logger_instance, extra or {})

    def log(self, level, msg, *args, **kwargs):
        if self.isEnabledFor(level):  # pragma: no cover
            msg, kwargs = self.process(msg, kwargs)
            self.logger._log(  # pylint: disable=protected-access
                level, LogBraceString(msg, args), (), **kwargs)


logger = LogStyleAdapter(  # pylint: disable=invalid-name
    logging.getLogger(__name__))


def makepath(path):
    """Convert relative path into absolute path."""
    return os.path.abspath(os.path.expanduser(path))


def list_all_files(path):
    """Get a list with relative paths for all files in the given path."""
    all_files = []
    for dirpath, dirs, filenames in os.walk(path):
        # Don't process hidden files
        dirs[:] = [d for d in dirs if not d[0] == '.']
        filenames = [f for f in filenames if not f[0] == '.']

        relpath = os.path.relpath(dirpath, path)
        for filename in filenames:
            all_files.append(os.path.normpath(os.path.join(relpath, filename)))
    return all_files


def ensure_directory_exists(path):
    """Ensure that the given path exists."""
    try:
        if not os.path.exists(path):
            os.makedirs(path)
    except OSError:
        pass


def delete_empty_directories(path):
    """Recursively remove empty directories."""
    if not os.path.isdir(path):
        return False
    if all([delete_empty_directories(os.path.join(path, filename))
            for filename in os.listdir(path)]):
        logger.info("Removing {}".format(path))
        os.rmdir(path)
        return True
    return False


def correct_path_fat32(filename):
    """Replace illegal characters in FAT32 filenames with '_'."""
    return re.sub(r'[\\|:|*|?|"|<|>|\|]', '_', filename)


def query_yes_no(question):  # pragma: no cover
    """Ask a yes/no question, yes being the default."""
    while 1:
        sys.stdout.write(question + ' [Y/n]: ')
        choice = input().lower()
        if choice == '':
            return True
        try:
            return choice.lower() in ['true', '1', 't', 'y', 'yes']
        except ValueError:
            sys.stdout.write("Valid answers: 'yes' or 'no'\n")



# sync_music - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""HashDb."""

import logging
import os
import pickle
import hashlib

from . import util

logger = util.LogStyleAdapter(  # pylint: disable=invalid-name
    logging.getLogger(__name__))


class HashDb:
    """Lightwight database for file hash values."""

    def __init__(self, path):
        self.database = {}
        self.path = path

    def load(self):
        """Load hash database to disk."""
        if os.path.exists(self.path):
            logger.info("Loading hash database from {}", self.path)
            hash_file = open(self.path, 'rb')
            self.database = pickle.load(hash_file, encoding="utf-8")
            hash_file.close()
        else:
            logger.info("No hash database file {}", self.path)

    def store(self):
        """Store hash database to disk."""
        logger.info("Storing hash database to {}", self.path)
        try:
            hash_file = open(self.path, 'wb')
            pickle.dump(self.database, hash_file)
            hash_file.close()
        except IOError:
            logger.error("Error: Failed to write hash database to {}",
                         self.path)

    @classmethod
    def get_hash(cls, path):
        """Calculate hash value for the given path."""
        hash_file = open(path, 'rb')
        hash_buffer = hash_file.read(4096)
        hash_file.close()
        return hashlib.md5(hash_buffer).hexdigest()



# music_sync - Sync music library to external device
# Copyright (C) 2013-2018 Christian Fetzer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Tests sync_music."""

import os
import shutil

from unittest import mock

import pytest

from sync_music.sync_music import SyncMusic
from sync_music.sync_music import load_settings
from sync_music.util import list_all_files


class TestSyncMusicSettings:
    """Tests sync_music load_settings function."""

    @staticmethod
    def test_noparams():
        """Tests loading of settings without parameters."""
        with pytest.raises(SystemExit):
            load_settings('')

    @staticmethod
    def test_nonexistent():
        """Tests loading of settings with incorrect paths."""
        argv = ['--audio-src', '/proc/nonexistingpath',
                '--audio-dest', '/proc/nonexistingpath']
        with pytest.raises(SystemExit):
            load_settings(argv)

    @staticmethod
    def test_forcecopy_with_hacks():
        """Tests loading of settings with force copy and hacks."""
        argv = ['--audio-src', '/tmp',
                '--audio-dest', '/tmp',
                '--mode=copy', '--albumartist-artist-hack']
        with pytest.raises(SystemExit):
            load_settings(argv)

    @staticmethod
    def test_configfile():
        """Tests loading of settings within config file."""
        filename = '/tmp/sync_music.cfg'
        argv = ['-c', filename]
        with open(filename, 'w') as configfile:
            configfile.write("[Defaults]\n")
            configfile.write("audio_src=/tmp\n")
            configfile.write("audio_dest=/tmp\n")
        try:
            load_settings(argv)
        finally:
            os.remove(filename)


class TestSyncMusicFiles():
    """Tests sync_music audio conversion."""

    input_path = 'tests/reference_data/regular'
    output_path = None

    output_files = [
        'stripped_flac.mp3', 'stripped_mp3.mp3', 'stripped_ogg.mp3',
        'withtags_flac.mp3', 'withtags_mp3.mp3', 'withtags_ogg.mp3',
        'sync_music.db', 'folder.jpg', 'dir/folder.jpg'
    ]

    @pytest.fixture(autouse=True)
    def init_output_path(self, tmpdir):
        """Setup temporary output directory."""
        self.output_path = str(tmpdir)

    def _execute_sync_music(self, input_path=input_path, output_files=None,
                            arguments=None, jobs=None):
        """Helper method to run sync_music tests."""
        if output_files is None:
            output_files = self.output_files
        argv = [] if arguments is None else arguments
        argv += ['--audio-src', input_path,
                 '--audio-dest', self.output_path]
        args = load_settings(argv)
        args.jobs = 1 if jobs is None else jobs
        sync_music = SyncMusic(args)
        sync_music.sync_audio()
        assert set(list_all_files(self.output_path)) == set(output_files)

    def test_emptyfolder(self):
        """Test empty input folder."""
        with pytest.raises(FileNotFoundError):
            self._execute_sync_music(self.output_path, [])

    def test_filenames_utf8(self):
        """Test UTF-8 input file names."""
        input_path = 'tests/reference_data/filenames_utf8'
        output_files = ['test_äöüß.mp3', 'sync_music.db']
        self._execute_sync_music(input_path, output_files)

    def test_filenames_fat32(self):
        """Test input file names incompatible with FAT32."""
        input_path = 'tests/reference_data/filenames_fat32'
        output_files = ['A___A.mp3', 'B___B.mp3', 'C___C.mp3', 'D___D.mp3',
                        'E___E.mp3', 'F___F.mp3', 'G___G.mp3', 'H___H.mp3',
                        'sync_music.db']
        self._execute_sync_music(input_path, output_files)

    def test_reference_default(self):
        """Test reference folder with default arguements."""
        self._execute_sync_music()

    def test_reference_withdatabase(self):
        """Test reference folder with default arguements (2 runs)."""
        self._execute_sync_music()
        self._execute_sync_music()

    def test_reference_cleanup(self):
        """Test reference folder with cleanup (file removed from input)."""
        input_path = 'tests/reference_data/regular_cleanup'
        output_files = self.output_files

        # Run a regular sync
        self._execute_sync_music()

        # First run: don't delete files
        def query_no(_):
            """Replacement for sync_music.util.query_yes_no."""
            return False
        with mock.patch('sync_music.util.query_yes_no', side_effect=query_no):
            self._execute_sync_music(input_path, output_files)

        # Delete a file also in output directory (to check double deletion)
        os.remove(os.path.join(self.output_path, 'withtags_ogg.mp3'))

        # Simulate OSError by creating a folder (that can't be removed)
        os.remove(os.path.join(self.output_path, 'withtags_mp3.mp3'))
        os.mkdir(os.path.join(self.output_path, 'withtags_mp3.mp3'))

        # Second run: delete files
        def query_yes(_):
            """Replacement for sync_music.util.query_yes_no."""
            return True
        output_files = [
            'stripped_mp3.mp3', 'sync_music.db'
        ]
        with mock.patch('sync_music.util.query_yes_no', side_effect=query_yes):
            self._execute_sync_music(input_path, output_files)

    def test_reference_multiprocessing(self):
        """Test reference folder with parallel jobs."""
        self._execute_sync_music(jobs=4)

    def test_reference_forcecopy(self):
        """Test reference folder with force copy."""
        output_files = [
            'stripped_flac.flac', 'stripped_mp3.mp3', 'stripped_ogg.ogg',
            'withtags_flac.flac', 'withtags_mp3.mp3', 'withtags_ogg.ogg',
            'sync_music.db', 'folder.jpg', 'dir/folder.jpg'
        ]
        self._execute_sync_music(output_files=output_files,
                                 arguments=['--mode=copy'])

    def test_reference_transcodeonly(self):
        """Test reference folder with tag processing only."""
        self._execute_sync_music()
        self._execute_sync_music(arguments=['-f', '--disable-file-processing'])

    def test_reference_tagsonly(self):
        """Test reference folder with file processing only."""
        self._execute_sync_music(arguments=['--disable-tag-processing'])

    def test_reference_hacks(self):
        """Test reference folder with hacks."""
        self._execute_sync_music(arguments=[
            '--albumartist-artist-hack', '--albumartist-composer-hack',
            '--discnumber-hack', '--tracknumber-hack'
        ])

    def test_reference_ioerror(self, mocker):
        """Test reference folder with mocked IOError."""
        mocker.patch('sync_music.transcode.Transcode.execute',
                     side_effect=IOError('Mocked exception'))
        mocker.patch('sync_music.actions.Copy.execute',
                     side_effect=IOError('Mocked exception'))
        self._execute_sync_music(output_files=['sync_music.db'])

    def test_reference_exception(self, mocker):
        """Test reference folder with mocked random exception."""
        mocker.patch('sync_music.transcode.Transcode.execute',
                     side_effect=Exception('Mocked exception'))
        mocker.patch('sync_music.actions.Copy.execute',
                     side_effect=Exception('Mocked exception'))
        self._execute_sync_music(output_files=['sync_music.db'])


class TestSyncMusicPlaylists():
    """Tests sync_music playlist conversion."""

    input_path = 'tests/reference_data/regular'
    output_path = None
    playlist_path = 'tests/reference_data/playlists'

    @pytest.fixture(autouse=True)
    def init_output_path(self, tmpdir):
        """Setup temporary output directory."""
        self.output_path = str(tmpdir)

    def _execute_sync_music(self, playlist_path=playlist_path):
        """Helper method to run sync_music tests."""
        argv = ['--audio-src', self.input_path,
                '--audio-dest', self.output_path,
                '--playlist-src', playlist_path]
        args = load_settings(argv)
        sync_music = SyncMusic(args)
        sync_music.sync_audio()
        sync_music.sync_playlists()

    def test_playlists(self):
        """Tests playlist generation."""
        self._execute_sync_music()

    def test_playlists_exists(self):
        """Tests playlist generation when playlist exists."""
        shutil.copy('tests/reference_data/playlists/normal.m3u',
                    os.path.join(self.output_path, 'normal.m3u'))
        self._execute_sync_music()

    def test_playlists_ioerror(self):
        """Tests playlist generation with playlist that can't be opened."""
        os.symlink('/dev/null', os.path.join(self.output_path, 'null.m3u'))
        self._execute_sync_music(self.output_path)
